"""
This is Model is for Decision Tree
"""

import numpy as np
from collections import Counter
from matplotlib import pyplot as plt
import ipywidgets as widgets
from tabulate import tabulate
from IPython import get_ipython
from IPython.display import display
from IPython.core.display import HTML
# from ..stats.stats import *
# from ..common import *

def euclidean_distance(a, b):
    return np.sqrt(np.sum((a - b) ** 2))


def log_loss(y_true, y_pred_probs, epsilon=1e-16):
    # safe_probs = np.where(y_pred_probs == 0, epsilon, y_pred_probs)
    # safe_probs = np.where(y_pred_probs == 1, 1 - epsilon, safe_probs)
    epsilon = np.finfo(y_pred_probs.dtype).eps
    y_pred_probs = np.clip(y_pred_probs, epsilon, 1 - epsilon)

    log_loss = -np.mean(y_true * np.log(y_pred_probs) + (1 - y_true) * np.log(1 - y_pred_probs))
    # log_loss_value = -(y_true * np.log(safe_probs) + (1 - y_true) * np.log(1 - safe_probs))
    return log_loss  # np.mean(log_loss_value)


def brier_score(y_true, y_pred_probs):
    return np.mean((y_true - y_pred_probs) ** 2)


class ConfusionMatrix:
    def __init__(self, y, y_pred, labels=None):
        """
        Initialize the Confusion Matrix for multiclass classification.

        Parameters:
        y (numpy array): True class labels.
        y_pred (numpy array): Predicted class labels.
        num_classes (int): Number of unique classes.
        """
        if labels is None:
            self.num_classes = len(np.unique(y))
            self.label_offset = np.min(y)  # Find the lowest label
        elif len(np.unique(y)) != len(labels):
            raise ValueError("Labels must have unique values")

        self.cm = np.zeros((self.num_classes, self.num_classes), dtype=int)

        # Populate the confusion matrix
        for true, pred in zip(y, y_pred):
            self.cm[true - self.label_offset, pred - self.label_offset] += 1
        # self.tp = np.sum((y_pred == 1) & (y == 1))
        # self.fp = np.sum((y_pred == 1) & (y == 0))
        # self.fn = np.sum((y_pred == 0) & (y == 1))
        # self.tn = np.sum((y_pred == 0) & (y == 0))

    def accuracy(self):
        """Calculate Accuracy score."""
        return np.trace(self.cm) / np.sum(self.cm)

    def precision(self, class_idx):
        """Calculate Precision for a specific class."""
        tp = self.cm[class_idx, class_idx]
        fp = np.sum(self.cm[:, class_idx]) - tp
        return tp / (tp + fp) if (tp + fp) != 0 else 0

    def recall(self, class_idx):
        """Calculate Recall for a specific class."""
        tp = self.cm[class_idx, class_idx]
        fn = np.sum(self.cm[class_idx, :]) - tp
        return tp / (tp + fn) if (tp + fn) != 0 else 0

    def f1_score(self, class_idx):
        """Calculate F1 score for a specific class."""
        p = self.precision(class_idx)
        r = self.recall(class_idx)
        return 2 * (p * r) / (p + r + 1e-8)

    def matthews_corrcoef(self):
        """Calculate Matthews Correlation Coefficient (MCC) for multiclass."""
        mcc = 0
        num_classes = self.num_classes
        for i in range(num_classes):
            tp = self.cm[i, i]
            fp = np.sum(self.cm[:, i]) - tp
            fn = np.sum(self.cm[i, :]) - tp
            tn = np.sum(self.cm) - tp - fp - fn

            numerator = (tp * tn) - (fp * fn)
            denominator = np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))
            mcc += numerator / (denominator + 1e-15)  # Avoid division by zero

        return mcc / num_classes  # Average MCC across all classes

    def cohen_kappa(self):
        """
        Calculate the Cohen's Kappa for multiclass classification.

        Parameters:
        y_true (numpy array): True class labels.
        y_pred (numpy array): Predicted class labels.
        num_classes (int): Number of unique classes.

        Returns:
        float: The Cohen's Kappa score.
        """
        # Initialize the confusion matrix
        total = np.sum(self.cm)

        # Calculate observed agreement (p_observed)
        p_observed = self.accuracy()

        # Calculate expected agreement (p_expected)
        row_sums = np.sum(self.cm, axis=1)
        col_sums = np.sum(self.cm, axis=0)
        p_expected = np.sum(row_sums * col_sums) / (total ** 2)

        # Calculate Cohen's Kappa
        kappa = (p_observed - p_expected) / (1 - p_expected)

        return kappa

    def overall_f1_score(self):
        """Calculate the overall F1 score as the macro-averaged F1 score."""
        f1_scores = [self.f1_score(i) for i in range(self.num_classes)]
        return np.mean(f1_scores)

    def classwise_metrics(self):
        """Return precision, recall, and F1 score for each class."""
        metrics = {}
        for i in range(self.num_classes):
            metrics[i] = {
                'precision': self.precision(i),
                'recall': self.recall(i),
                'f1_score': self.f1_score(i)
            }
        return metrics


def roc_auc(y_true, y_pred_probs):
    thresholds = np.sort(y_pred_probs)[::-1]
    tpr_list = []
    fpr_list = []

    for threshold in thresholds:
        predictions = (y_pred_probs >= threshold).astype(int)
        tp = np.sum((predictions == 1) & (y_true == 1))
        fp = np.sum((predictions == 1) & (y_true == 0))
        tn = np.sum((predictions == 0) & (y_true == 0))
        fn = np.sum((predictions == 0) & (y_true == 1))

        tpr = tp / (tp + fn)
        fpr = fp / (fp + tn)
        tpr_list.append(tpr)
        fpr_list.append(fpr)

    auc = 0.0
    for i in range(1, len(tpr_list)):
        auc += (fpr_list[i] - fpr_list[i - 1]) * (tpr_list[i] + tpr_list[i - 1]) / 2

    return auc


def display_table_in_jupyter(metrics_dict):
    try:
        metrics_df = pd.DataFrame(metrics_dict)
        styled_df = metrics_df.style.format({
            "Value": "{:.4f}"  # Adjust format as needed
        }).set_table_styles([
            {'selector': 'thead th',
             'props': [('background-color', '#f7f7f9'), ('color', '#333'), ('font-weight', 'bold')]},
            {'selector': 'tbody td', 'props': [('text-align', 'center')]},
        ]).set_properties(**{
            'background-color': '#f4f4f9',  # Light background for table cells
            'border': '1px solid #ddd',  # Border for table cells
        })
        display(styled_df)

    except Exception:
        print(tabulate(metrics_dict.items(), headers=["Evaluation Metric", "Value"], tablefmt="grid"))


class Node:
    def __init__(self, feature=None, threshold=None, left=None, right=None, *, value=None):
        self.feature = feature
        self.threshold = threshold
        self.left = left
        self.right = right
        self.value = value

    def is_leaf_node(self):
        return self.value is not None


class DecisionTree:
    def __init__(self, min_samples_split=2, max_depth=100, n_features=None, normalize=True):
        self.min_samples_split = min_samples_split
        self.max_depth = max_depth
        self.n_features = n_features
        self.normalize = normalize
        self.root = None
        self.labels = None

    def fit(self, X, y, split: int = 0.7, shuffle=False, seed=None):
        """
        Initializes the training environment.

        :param X: numpy array
            Contains independent variables.
        :param y: numpy array
                Contains dependent variables.
        """
        if not isinstance(X, np.ndarray) or not isinstance(y, np.ndarray):
            raise ValueError("X and y must be numpy arrays.")
        self.n_features = X.shape[1] if not self.n_features else min(X.shape[1], self.n_features)
        self.root = self._grow_tree(X, y)
        self.labels = np.unique(y)

    def _grow_tree(self, X, y, depth=0):
        n_samples, n_feats = X.shape
        n_labels = len(np.unique(y))

        if (depth >= self.max_depth or n_labels == 1 or n_samples < self.min_samples_split):
            leaf_value = self._most_common_label(y)
            return Node(value=leaf_value)
        feat_idxs = np.random.choice(n_feats, self.n_features, replace=False)
        best_feature, best_thres = self._best_split(X, y, feat_idxs)

        left_idxs, right_idxs = self._split(X[:, best_feature], best_thres)

        current_depth = depth + 1
        left = self._grow_tree(X[left_idxs, :], y[left_idxs], depth=current_depth)
        right = self._grow_tree(X[right_idxs, :], y[right_idxs], depth=current_depth)
        return Node(best_feature, best_thres, left, right)

    def _best_split(self, X, y, feat_idxs):
        best_gain = -1
        split_idx, split_threshold = None, None

        for feat_idx in feat_idxs:
            X_column = X[:, feat_idx]
            thresholds = np.unique(X_column)
            for threshold in thresholds:
                gain = self._information_gain(y, X_column, threshold)

                if gain > best_gain:
                    best_gain = gain
                    split_idx = feat_idx
                    split_threshold = threshold
        return split_idx, split_threshold

    def _information_gain(self, y, X_column, threshold):
        """IG=parent entropy - sum (weight*entropy of children)"""
        parent_entropy = self._entropy(y)

        left_idxs, right_idxs = self._split(X_column, threshold)

        if len(left_idxs) == 0 or len(right_idxs) == 0:
            return 0
        n = len(y)
        n_l, n_r = len(left_idxs), len(right_idxs)

        e_l, e_r = self._entropy(y[left_idxs]), self._entropy(y[right_idxs])
        child_entropy = (n_l / n) * e_l + (n_r / n) * e_r
        ig = parent_entropy - child_entropy
        return ig

    def _split(self, X_column, split_thresh):
        left_idxs = np.argwhere(X_column <= split_thresh).flatten()
        right_idxs = np.argwhere(X_column > split_thresh).flatten()
        return left_idxs, right_idxs

    def _entropy(self, y):
        hist = np.bincount(y)
        ps = hist / len(y)
        return -np.sum([p * np.log(p) for p in ps if p > 0])

    def _most_common_label(self, y):
        counter = Counter(y)
        value = counter.most_common(1)[0][0]
        return value

    def predict(self, X):
        """
        Summary of the function.

        Args:
            param1 (int): Description of param1.
            param2 (str): Description of param2.

        Returns:
            bool: Description of return value.
        """
        return np.array([self._traverse_tree(x, self.root) for x in X])

    def _traverse_tree(self, x, node):
        if node.is_leaf_node():
            return node.value

        if x[node.feature] <= node.threshold:
            return self._traverse_tree(x, node.left)
        return self._traverse_tree(x, node.right)


    def evaluate(self, X=None, y=None):
        """
        Evaluate model performance on test data.

        Parameters:
        X (numpy.ndarray): Test data of shape (n_samples, n_features).
        y (numpy.ndarray): True labels for X.

        Returns:
        dict: Dictionary with accuracy, precision, recall, and F1 score.
        """
        if X is None or y is None:
            X = self.X_train
            y = self.y_train

        if len(y) != X.shape[0]:
            raise ValueError("Dimension mismatch between X and y.")

        y_pred = np.asarray(self.predict(X))
        confusion_matrix = ConfusionMatrix(y, y_pred)
        accuracy_value = confusion_matrix.accuracy()
        # precision_value = confusion_matrix.precision()
        # recall_value = confusion_matrix.recall()
        p_r_f1 = confusion_matrix.classwise_metrics()
        f1_score_value = confusion_matrix.overall_f1_score()
        # log_loss_value = log_loss(y, y_pred)
        cohen_kappa_value = confusion_matrix.cohen_kappa()
        matthews_corrcoef_value = confusion_matrix.matthews_corrcoef()
        # brier_score_value = brier_score(y, y_pred)

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            metrics_dict = {
                "Evaluation Metric": ["Accuracy", "Precision, Recall and F1", "F1 (Macro Avg)",
                                      # "Log Loss",
                                      "Cohen Kappa",
                                      "Matthews Correlation Coefficient",
                                      # "Brier Score"
                                      ],
                "Value": [accuracy_value, p_r_f1, f1_score_value,
                          # log_loss_value,
                          cohen_kappa_value, matthews_corrcoef_value,
                          # brier_score_value
                          ],
            }
            display_table_in_jupyter(metrics_dict)

        else:
            p_r_ = []
            for class_idx, metric in p_r_f1.items():
                p_r_.append([self.labels[class_idx], metric['precision'], metric['recall'], metric['f1_score']])
            metrics_dict = {
                "Accuracy": accuracy_value,
                "Precision, Recall and F1": tabulate(p_r_, headers=["Class", "Precision", "Recall", "F1"],
                                                     tablefmt="pretty"),
                "F1 (Macro Avg)": f1_score_value,
                # "Log Loss": log_loss_value,
                "Cohen Kappa": cohen_kappa_value,
                "Matthews Correlation Coefficient": matthews_corrcoef_value,
                # "Brier Score": brier_score_value,
            }
            print(tabulate(metrics_dict.items(), headers=["Evaluation Metric", "Value"], tablefmt="grid"))


__all__ = ['DecisionTree']
