import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from tabulate import tabulate
import ipywidgets as widgets
from IPython import get_ipython
from IPython.display import display
from IPython.core.display import HTML
from ..stats.stats import *
from ..common import *
from collections import Counter


class KNN:
    def __init__(self, k=3, labels=None):
        self.k = k
        self.labels = labels

    def fit(self, X, y):
        self.X_train = X
        self.y_train = y

    def predict(self, X):
        predictions = [self._predict(x) for x in X]
        return predictions

    def _predict(self, x):
        # Calculate distance, euclidean here
        distances = [euclidean_distance(x, x_train) for x_train in self.X_train]

        k_indices = np.argsort(distances)[:self.k]
        k_nearest_labels = [self.y_train[i] for i in k_indices]

        most_common = Counter(k_nearest_labels).most_common()
        return most_common[0][0]

    def evaluate(self, X=None, y=None):
        """
        Evaluate model performance on test data.

        Parameters:
        X (numpy.ndarray): Test data of shape (n_samples, n_features).
        y (numpy.ndarray): True labels for X.

        Returns:
        dict: Dictionary with accuracy, precision, recall, and F1 score.
        """
        if X is None or y is None:
            X = self.X_train
            y = self.y_train

        if len(y) != X.shape[0]:
            raise ValueError("Dimension mismatch between X and y.")

        y_pred = np.asarray(self.predict(X))
        confusion_matrix = ConfusionMatrix(y, y_pred)
        accuracy_value = confusion_matrix.accuracy()
        # precision_value = confusion_matrix.precision()
        # recall_value = confusion_matrix.recall()
        p_r_f1 = confusion_matrix.classwise_metrics()
        f1_score_value = confusion_matrix.overall_f1_score()
        # log_loss_value = log_loss(y, y_pred)
        cohen_kappa_value = confusion_matrix.cohen_kappa()
        matthews_corrcoef_value = confusion_matrix.matthews_corrcoef()
        # brier_score_value = brier_score(y, y_pred)

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            metrics_dict = {
                "Evaluation Metric": ["Accuracy", "Precision, Recall and F1", "F1 (Macro Avg)",
                                      # "Log Loss",
                                      "Cohen Kappa",
                                      "Matthews Correlation Coefficient",
                                      # "Brier Score"
                                      ],
                "Value": [accuracy_value, p_r_f1, f1_score_value,
                          # log_loss_value,
                          cohen_kappa_value, matthews_corrcoef_value,
                          # brier_score_value
                          ],
            }
            display_table_in_jupyter(metrics_dict)

        else:
            p_r_ = []
            for class_idx, metric in p_r_f1.items():
                p_r_.append([self.labels[class_idx], metric['precision'], metric['recall'], metric['f1_score']])
            metrics_dict = {
                "Accuracy": accuracy_value,
                "Precision, Recall and F1": tabulate(p_r_, headers=["Class", "Precision", "Recall", "F1"],
                                                     tablefmt="pretty"),
                "F1 (Macro Avg)": f1_score_value,
                # "Log Loss": log_loss_value,
                "Cohen Kappa": cohen_kappa_value,
                "Matthews Correlation Coefficient": matthews_corrcoef_value,
                # "Brier Score": brier_score_value,
            }
            print(tabulate(metrics_dict.items(), headers=["Evaluation Metric", "Value"], tablefmt="grid"))


__all__ = ['KNN']
