import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from tabulate import tabulate
import ipywidgets as widgets
from IPython import get_ipython
from IPython.display import display
from IPython.core.display import HTML
from ..stats.stats import *
from ..common import *


class LogisticRegression:
    """
    Logistic Regression classifier using gradient descent optimization.
    """

    def __init__(self, lr=0.001, n_iters=1000, threshold=0.5):
        """
        Parameters:
        lr (float): Learning rate (default is 0.001).
        n_iters (int): Number of training iterations (default is 1000).
        """
        self.lr = lr
        self.n_iters = n_iters
        self.threshold = threshold
        self.weights = None
        self.bias = None
        self.X = None
        self.y = None
        self.y_pred_probs = None

    def fit(self, X, y):
        """
        Train the Logistic Regression model using gradient descent.

        Parameters:
        X (numpy.ndarray): Training data of shape (n_samples, n_features).
        y (numpy.ndarray): Target values of shape (n_samples,).

        """
        self.X = X
        self.y = y
        n_samples, n_features = X.shape
        self.weights = np.zeros(n_features)
        self.bias = 0

        for _ in range(self.n_iters):
            linear_pred = np.dot(X, self.weights) + self.bias
            predictions = sigmoid(linear_pred)

            dw = (1 / n_samples) * np.dot(X.T, (predictions - y))
            db = (1 / n_samples) * np.sum(predictions - y)

            self.weights -= self.lr * dw
            self.bias -= self.lr * db

    def predict(self, X):
        """
        Predict binary class labels for samples in X.

        Parameters:
        X (numpy.ndarray): Input data of shape (n_samples, n_features).

        Returns:
        list: Predicted class labels (0 or 1) for each sample.
        """
        linear_pred = np.dot(X, self.weights) + self.bias
        y_pred = sigmoid(linear_pred)
        self.y_pred_probs = y_pred
        predictions = (y_pred >= self.threshold).astype(int)
        return predictions

    def predict_probs(self, X):
        """
        Predict binary class labels for samples in X.

        Parameters:
        X (numpy.ndarray): Input data of shape (n_samples, n_features).

        Returns:
        list: Predicted probabilities for class label for each sample.
        """
        linear_pred = np.dot(X, self.weights) + self.bias
        return sigmoid(linear_pred)

    def evaluate(self, X=None, y=None):
        """
        Evaluate model performance on test data.

        Parameters:
        X (numpy.ndarray): Test data of shape (n_samples, n_features).
        y (numpy.ndarray): True labels for X.

        Returns:
        dict: Dictionary with accuracy, precision, recall, and F1 score.
        """
        if X is None or y is None:
            X = self.X
            y = self.y

        if len(y) != X.shape[0]:
            raise ValueError("Dimension mismatch between X and y.")

        y_pred = np.asarray(self.predict(X))
        confusion_matrix = ConfusionMatrix(y, y_pred)
        accuracy_value = confusion_matrix.accuracy()
        precision_value = confusion_matrix.precision()
        recall_value = confusion_matrix.recall()
        f1_score_value = confusion_matrix.f1_score()
        log_loss_value = log_loss(y, y_pred)
        cohen_kappa_value = cohen_kappa(y, y_pred)
        matthews_corrcoef_value = confusion_matrix.matthews_corrcoef()
        brier_score_value = brier_score(y, y_pred)

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            metrics_dict = {
                "Evaluation Metric": ["Accuracy", "Precision", "Recall", "F1",
                                      # "Log Loss",
                                      "Cohen Kappa",
                                      "Matthews Correlation Coefficient", "Brier Score"],
                "Value": [accuracy_value, precision_value, recall_value, f1_score_value,
                          # log_loss_value,
                          cohen_kappa_value, matthews_corrcoef_value, brier_score_value],
            }
            display_table_in_jupyter(metrics_dict)

        else:
            metrics_dict = {
                "Accuracy": accuracy_value,
                "Precision": precision_value,
                "Recall": recall_value,
                "F1": f1_score_value,
                # "Log Loss": log_loss_value,
                "Cohen Kappa": cohen_kappa_value,
                "Matthews Correlation Coefficient": matthews_corrcoef_value,
                "Brier Score": brier_score_value,
            }
            print(tabulate(metrics_dict.items(), headers=["Evaluation Metric", "Value"], tablefmt="grid"))

    def plot(self, X=None, y=None):
        if X is None or y is None:
            X = self.X
            y = self.y
        y_pred_probs = self.predict_probs(X)
        y_pred = self.predict(X)

        def plot1():
            plt.figure(figsize=(12, 8))
            thresholds = np.sort(y_pred_probs)[::-1]
            tpr_list = []
            fpr_list = []

            for threshold in thresholds:
                # Calculate TP, FP, TN, FN
                y_pred = (y_pred_probs >= threshold).astype(int)
                TP = np.sum((y_pred == 1) & (y == 1))
                FP = np.sum((y_pred == 1) & (y == 0))
                TN = np.sum((y_pred == 0) & (y == 0))
                FN = np.sum((y_pred == 0) & (y == 1))

                # Calculate TPR and FPR
                TPR = TP / (TP + FN) if (TP + FN) > 0 else 0
                FPR = FP / (FP + TN) if (FP + TN) > 0 else 0

                tpr_list.append(TPR)
                fpr_list.append(FPR)

            # Plot ROC curve
            plt.figure()
            plt.plot(fpr_list, tpr_list, color="blue", label="ROC Curve (scratch)")
            plt.plot([0, 1], [0, 1], "k--", label="Random Classifier")
            plt.xlabel("False Positive Rate")
            plt.ylabel("True Positive Rate")
            plt.title("ROC Curve (From Scratch)")
            plt.legend()
            plt.show()

        def plot2():
            plt.figure(figsize=(12, 8))
            cm = ConfusionMatrix(y, y_pred)
            cm = np.array([[cm.tn, cm.fp], [cm.fn, cm.tp]])

            # Plot heatmap
            plt.figure()
            sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", xticklabels=["Pred 0", "Pred 1"],
                        yticklabels=["True 0", "True 1"])
            plt.xlabel("Predicted Label")
            plt.ylabel("True Label")
            plt.title("Confusion Matrix (From Scratch)")
            plt.show()

        def plot3():
            thresholds = np.sort(y_pred_probs)[::-1]
            precision_list = []
            recall_list = []

            for threshold in thresholds:
                # Calculate TP, FP, FN
                y_pred = (y_pred_probs >= threshold).astype(int)
                TP = np.sum((y_pred == 1) & (y == 1))
                FP = np.sum((y_pred == 1) & (y == 0))
                FN = np.sum((y_pred == 0) & (y == 1))

                # Calculate precision and recall
                precision = TP / (TP + FP) if (TP + FP) > 0 else 0
                recall = TP / (TP + FN) if (TP + FN) > 0 else 0

                precision_list.append(precision)
                recall_list.append(recall)

            # Plot Precision-Recall curve
            plt.figure()
            plt.plot(recall_list, precision_list, color="purple", label="Precision-Recall Curve (scratch)")
            plt.xlabel("Recall")
            plt.ylabel("Precision")
            plt.title("Precision-Recall Curve (From Scratch)")
            plt.legend()
            plt.show()

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            custom_style = """
                            <style>
                                .widget-tab-contents {
                                    border: 2px solid #4CAF50;
                                    border-radius: 5px;
                                    padding: 10px;
                                }
                                .widget-tab {
                                    background-color: #f2f2f2;
                                }
                                .widget-tab > .p-TabBar-tab {
                                    background-color: #4CAF50;
                                    color: white;
                                    padding: 10px;
                                    border-radius: 4px;
                                }
                                .widget-tab > .p-TabBar-tab.p-mod-selected {
                                    background-color: #45a049;
                                    font-weight: bold;
                                }
                                .widget-tab > .p-TabBar-tab:hover {
                                    background-color: #66bb6a;
                                }
                            </style>
                            """
            display(HTML(custom_style))
            tab1 = widgets.Output()
            tab2 = widgets.Output()
            tab3 = widgets.Output()
            tab4 = widgets.Output()
            tab5 = widgets.Output()
            tab6 = widgets.Output()

            with tab1:
                plot1()
            with tab2:
                plot2()
            with tab3:
                plot3()
            # with tab4:
            #     plot4()
            # with tab5:
            #     plot5()
            # with tab6:
            #     plot6()

            tab_widget = widgets.Tab(children=[tab1, tab2, tab3, tab4, tab5, tab6])
            tab_widget.set_title(0, 'Confusion Matrix')
            tab_widget.set_title(1, 'Silhouette Plot')
            tab_widget.set_title(2, 'Precision-Recall Curve')
            # tab_widget.set_title(3, 'Q-Q Plot')

            display(tab_widget)
        else:
            plot1()
            plot2()
            plot3()
            # plot4()
            # plot5()
            # plot6()


__all__ = ['LogisticRegression']
