import numpy as np
from matplotlib import pyplot as plt
import ipywidgets as widgets
from tabulate import tabulate
from IPython import get_ipython
from IPython.display import display
from IPython.core.display import HTML
from ..stats.stats import *
from ..common import *


class KMeans:
    def __init__(self, K=5, max_iters=100, plot_steps=False):
        self.K = K
        self.max_iters = max_iters
        self.plot_steps = plot_steps
        self.clusters = [[] for _ in range(self.K)]
        self.centroids = []
        self.labels = None

    def _get_cluster_labels(self, clusters):
        labels = np.empty(self.n_samples)
        for cluster_idx, cluster in enumerate(clusters):
            for sample_idx in cluster:
                labels[sample_idx] = cluster_idx
        return labels

    def _create_clusters(self, centroids):
        clusters = [[] for _ in range(self.K)]
        for idx, sample in enumerate(self.X):
            centroid_idx = self._closest_centroid(sample, centroids)
            clusters[centroid_idx].append(idx)
        return clusters

    def _closest_centroid(self, sample, centroids):
        distances = [euclidean_distance(sample, point) for point in centroids]
        closest_idx = np.argmin(distances)
        return closest_idx

    def _get_centroids(self, clusters):
        centroids = np.zeros((self.K, self.n_features))
        for cluster_idx, cluster in enumerate(clusters):
            cluster_mean = np.mean(self.X[cluster], axis=0)
            centroids[cluster_idx] = cluster_mean
        return centroids

    def _is_converged(self, centroids_old, centroids):
        distances = [euclidean_distance(centroids_old[i], centroids[i]) for i in range(self.K)]
        return sum(distances) == 0

    def fit(self, X):
        """
        Fit the model to the training data.

        X: numpy array; Contains independent variables.
        """
        self.X = X
        self.n_samples, self.n_features = X.shape

        random_sample_idxs = np.random.choice(self.n_samples, self.K, replace=False)
        self.centroids = [self.X[idx] for idx in random_sample_idxs]

        for _ in range(self.max_iters):
            self.clusters = self._create_clusters(self.centroids)

            if self.plot_steps:
                self.plot()

            centroids_old = self.centroids
            self.centroids = self._get_centroids(self.clusters)

            if self._is_converged(centroids_old, self.centroids):
                break
            if self.plot_steps:
                self.plot()
        self.labels = self._get_cluster_labels(self.clusters)
        return self.labels

    def plot(self):
        """Plot the results of the KMeans Clustering."""
        if self.X.shape[1] != 2:
            print("Plot not possible")
            return

        def plot1():
            plt.figure(figsize=(12, 8))
            for i, index in enumerate(self.clusters):
                point = self.X[index].T
                plt.scatter(*point)

            for point in self.centroids:
                plt.scatter(*point, marker="x", color="black", linewidth=2)

            plt.show()

        def plot2():
            plt.figure(figsize=(12, 8))
            y_lower = 10
            silhouette_vals = silhouette_values(self.X, self.centroids, self.labels)

            for i in range(self.K):
                print(self.K, i)
                cluster_silhouette_vals = [x for idx, x in enumerate(silhouette_vals) if
                                           self.labels[idx] == i]
                cluster_silhouette_vals.sort()

                y_upper = y_lower + len(cluster_silhouette_vals)

                plt.fill_betweenx(np.arange(y_lower, y_upper), 0, cluster_silhouette_vals, alpha=0.7)

                plt.text(-0.05, (y_lower + y_upper) / 2, str(i))

                y_lower = y_upper + 10

            silhouette_score = np.mean(silhouette_vals)
            plt.axvline(x=silhouette_score, color="blue", linestyle="--",
                        label='Average Silhouette Score: {:.4f}'.format(silhouette_score))
            plt.legend(loc='best')
            plt.title("Silhouette Plot")
            plt.xlabel("Silhouette Coefficient Values")
            plt.ylabel("Cluster Label")
            plt.xlim([-0.1, 1])

            plt.show()

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            custom_style = """
                    <style>
                        .widget-tab-contents {
                            border: 2px solid #4CAF50;
                            border-radius: 5px;
                            padding: 10px;
                        }
                        .widget-tab {
                            background-color: #f2f2f2;
                        }
                        .widget-tab > .p-TabBar-tab {
                            background-color: #4CAF50;
                            color: white;
                            padding: 10px;
                            border-radius: 4px;
                        }
                        .widget-tab > .p-TabBar-tab.p-mod-selected {
                            background-color: #45a049;
                            font-weight: bold;
                        }
                        .widget-tab > .p-TabBar-tab:hover {
                            background-color: #66bb6a;
                        }
                    </style>
                    """
            display(HTML(custom_style))
            tab1 = widgets.Output()
            tab2 = widgets.Output()
            tab3 = widgets.Output()
            tab4 = widgets.Output()
            tab5 = widgets.Output()
            tab6 = widgets.Output()

            with tab1:
                plot1()
            with tab2:
                plot2()
            # with tab3:
            #     plot3()
            # with tab4:
            #     plot4()
            # with tab5:
            #     plot5()
            # with tab6:
            #     plot6()

            tab_widget = widgets.Tab(children=[tab1, tab2, tab3, tab4, tab5, tab6])
            tab_widget.set_title(0, 'Cluster Plot')
            tab_widget.set_title(1, 'Silhouette Plot')
            # tab_widget.set_title(2, 'Histogram')
            # tab_widget.set_title(3, 'Q-Q Plot')
            # tab_widget.set_title(4, 'Model Performance Curve')
            # tab_widget.set_title(5, 'Learning Curve')

            display(tab_widget)
        else:
            plot1()
            plot2()
            # plot3()
            # plot4()
            # plot5()
            # plot6()

    def evaluate(self, X=None):
        """Evaluate the model on the test set."""
        inertia_value = inertia(X, self.centroids, self.labels)
        silhouette_score_value = silhouette_score(X, self.centroids, self.labels)
        calinski_harabasz_index_value = calinski_harabasz_index(X, self.labels)
        davies_bouldin_index_value = davies_bouldin_index(X, self.labels)

        shell_name = get_ipython().__class__.__name__
        if shell_name == 'ZMQInteractiveShell' or shell_name == 'Shell':
            metrics_dict = {
                "Evaluation Metric": ["Inertia", "Silhouette Score", "Calinski-Harabasz Index", "Davies-Bouldin Index"],
                "Value": [inertia_value, silhouette_score_value, calinski_harabasz_index_value,
                          davies_bouldin_index_value],
            }
            display_table_in_jupyter(metrics_dict)

        else:
            metrics_dict = {
                "Inertia": inertia_value,
                "Silhouette Score": silhouette_score_value,
                "Calinski-Harabasz Index": calinski_harabasz_index_value,
                "Davies-Bouldin Index": davies_bouldin_index_value
            }
            print(tabulate(metrics_dict.items(), headers=["Evaluation Metric", "Value"], tablefmt="grid"))


__all__ = ['KMeans']
