import numpy as np


def euclidean_distance(a, b):
    return np.sqrt(np.sum((a - b) ** 2))


def inertia(X, centroids, labels):
    inertia = 0.0
    for i in range(X.shape[0]):
        centroid = centroids[labels[i]]
        distance = euclidean_distance(X[i], centroid)
        inertia += distance ** 2
    return inertia


def silhouette_score(X, c, labels):
    try:
        n_samples = X.shape[0]
        unique_labels = set(labels)
        silhouette_scores = []

        for i in range(n_samples):
            current_label = labels[i]
            same_cluster_points = X[labels == current_label]
            if len(same_cluster_points) > 1:
                a = np.mean([euclidean_distance(X[i], point) for point in same_cluster_points if
                             not np.array_equal(X[i], point)])
            else:
                a = 0

            b = float('inf')
            for other_label in unique_labels:
                if other_label != current_label:
                    other_cluster_points = X[labels == other_label]
                    mean_distance_to_other_cluster = np.mean(
                        [euclidean_distance(X[i], point) for point in other_cluster_points])
                    b = min(b, mean_distance_to_other_cluster)

            silhouette_score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            silhouette_scores.append(silhouette_score)

        return np.mean(silhouette_scores)
    except Exception as e:
        print(f"Error calculating silhouette score: {e}")
        return None


def silhouette_values(X, c, labels):
    try:
        n_samples = X.shape[0]
        unique_labels = set(labels)
        silhouette_scores = []

        for i in range(n_samples):
            current_label = labels[i]
            same_cluster_points = X[labels == current_label]
            if len(same_cluster_points) > 1:
                a = np.mean([euclidean_distance(X[i], point) for point in same_cluster_points if
                             not np.array_equal(X[i], point)])
            else:
                a = 0

            b = float('inf')
            for other_label in unique_labels:
                if other_label != current_label:
                    other_cluster_points = X[labels == other_label]
                    mean_distance_to_other_cluster = np.mean(
                        [euclidean_distance(X[i], point) for point in other_cluster_points])
                    b = min(b, mean_distance_to_other_cluster)

            silhouette_score = (b - a) / max(a, b) if max(a, b) != 0 else 0
            silhouette_scores.append(silhouette_score)

        return silhouette_scores
    except Exception as e:
        print(f"Error calculating silhouette score: {e}")
        return None


def calinski_harabasz_index(X, labels):
    n = X.shape[0]
    k = len(np.unique(labels))

    if k < 2:
        return 0
    overall_mean = np.mean(X, axis=0)

    B = sum(len(X[labels == i]) * np.linalg.norm(np.mean(X[labels == i], axis=0) - overall_mean) ** 2 for i in
            np.unique(labels))

    W = sum(np.sum((X[labels == i] - np.mean(X[labels == i], axis=0)) ** 2) for i in np.unique(labels))

    return (B / W) * ((n - k) / (k - 1))


def davies_bouldin_index(X, labels):
    unique_labels = np.unique(labels)
    k = len(unique_labels)

    if k < 2:
        return float('inf')

    centroids = np.array([np.mean(X[labels == label], axis=0) for label in unique_labels])
    intra_cluster_distances = np.array(
        [np.mean(np.linalg.norm(X[labels == label] - centroid, axis=1)) for label, centroid in
         zip(unique_labels, centroids)])

    DB = 0.0
    for i in range(k):
        max_ratio = 0.0
        for j in range(k):
            if i != j:
                distance_between_centroids = np.linalg.norm(centroids[i] - centroids[j])
                ratio = (intra_cluster_distances[i] + intra_cluster_distances[j]) / distance_between_centroids
                max_ratio = max(max_ratio, ratio)
        DB += max_ratio

    return DB / k


def mae(a, b):
    """Calculate Mean Absolute Error."""
    return np.mean(np.abs(a - b))


def mse(a, b):
    """Calculate Mean Squared Error."""
    return np.mean((a - b) ** 2)


def rmse(a, b):
    """Calculate Root Mean Squared Error."""
    return np.sqrt(np.mean((a - b) ** 2))


def rmsle(a, b):
    """Calculate Root Mean Squared Log Error."""
    return np.sqrt(np.mean((np.log1p(a) - np.log1p(b)) ** 2))


def mape(a, b):
    """Calculate Mean Absolute Percentage Error."""
    return np.mean(np.abs((a - b) / a)) * 100


def correlation(a, b):
    """Calculate correlation coefficient."""
    cov_matrix = np.cov(a, b)
    return cov_matrix[0, 1] / np.sqrt(cov_matrix[0, 0] * cov_matrix[1, 1])


def sigmoid(x):
    """
    Compute the sigmoid function.

    Parameters:
    x (array-like): Input array or scalar.

    Returns:
    numpy.ndarray: Sigmoid of input x.

    Raises:
    ValueError: If x contains non-numeric values.
    OverflowError: If x values are too large and cause overflow.
    """
    try:
        x = np.clip(x, -700, float('inf'))

        return 1 / (1 + np.exp(-x))
    except OverflowError:
        raise OverflowError("Input values in 'x' are too large, causing overflow in exponential calculation.")
    except TypeError:
        raise ValueError("Non-numeric values found in input 'x'. Sigmoid function requires numeric input.")


def log_loss(y_true, y_pred_probs, epsilon=1e-16):
    y_pred_probs = np.clip(y_pred_probs, epsilon, 1 - epsilon)
    log_loss = -np.mean(y_true * np.log(y_pred_probs) + (1 - y_true) * np.log(1 - y_pred_probs))
    return log_loss


def cohen_kappa(y_true, y_pred):
    total = len(y_true)
    p_observed = np.sum(y_true == y_pred) / total
    p_expected = (np.sum(y_true) * np.sum(y_pred) + np.sum(1 - y_true) * np.sum(1 - y_pred)) / total ** 2
    kappa = (p_observed - p_expected) / (1 - p_expected)
    return kappa


def brier_score(y_true, y_pred_probs):
    return np.mean((y_true - y_pred_probs) ** 2)


def roc_auc(y_true, y_pred_probs):
    thresholds = np.sort(y_pred_probs)[::-1]
    tpr_list = []
    fpr_list = []

    for threshold in thresholds:
        predictions = (y_pred_probs >= threshold).astype(int)
        tp = np.sum((predictions == 1) & (y_true == 1))
        fp = np.sum((predictions == 1) & (y_true == 0))
        tn = np.sum((predictions == 0) & (y_true == 0))
        fn = np.sum((predictions == 0) & (y_true == 1))

        tpr = tp / (tp + fn)
        fpr = fp / (fp + tn)
        tpr_list.append(tpr)
        fpr_list.append(fpr)

    auc = 0.0
    for i in range(1, len(tpr_list)):
        auc += (fpr_list[i] - fpr_list[i - 1]) * (tpr_list[i] + tpr_list[i - 1]) / 2

    return auc


class ConfusionMatrix:
    def __init__(self, y, y_pred):
        """
        Initialize the Logistic Regression model.

        Parameters:
        y (numpy array): .
        y_pred (numpy array): .
        """
        self.tp = np.sum((y_pred == 1) & (y == 1))
        self.fp = np.sum((y_pred == 1) & (y == 0))
        self.fn = np.sum((y_pred == 0) & (y == 1))
        self.tn = np.sum((y_pred == 0) & (y == 0))

    def accuracy(self):
        """Calculate Accuracy score."""
        return (self.tp + self.tn) / (self.tp + self.fp + self.fn + self.tn)

    def precision(self):
        """Calculate Precision score."""
        return self.tp / (self.tp + self.fp) if (self.tp + self.fp) != 0 else 0

    def recall(self):
        """Calculate Recall score."""
        return self.tp / (self.tp + self.fn) if (self.tp + self.fn) != 0 else 0

    def f1_score(self):
        """Calculate F1 score."""
        p = self.precision()
        r = self.recall()
        return 2 * (p * r) / (p + r)

    def matthews_corrcoef(self):
        """Calculate Matthews Correlation Coefficient (MCC) score."""
        numerator = (self.tp * self.tn) - (self.fp * self.fn)
        denominator = np.sqrt((self.tp + self.fp) * (self.tp + self.fn) * (self.tn + self.fp) * (self.tn + self.fn))
        mcc = numerator / (denominator + 1e-15)  # Add epsilon to avoid division by zero
        return mcc
